#include "fw.h"
#include "console.h"
#include "io.h"

extern char __dram_init_start[];
abssymval(_dram_init_len)
extern char __dram_bss_start[];
abssymval(_dram_bss_len)

enum romcmd {
    ROM_WRITE_ENABLE			= 0x06,
    ROM_VOLATILE_SR_WRITE_ENABLE	= 0x50,
    ROM_WRITE_DISABLE			= 0x04,
    ROM_RELEASE_POWERDOWN_ID		= 0xab,
    ROM_MANUFACTURER_DEVICE_ID		= 0x90,
    ROM_JEDEC_ID			= 0x9f,
    ROM_READ_UNIQUE_ID			= 0x4b,
    ROM_READ_DATA			= 0x03, /* DO NOT USE */
    ROM_FAST_READ			= 0x0b, /* Only for non-DMA */
    ROM_PAGE_PROGRAM			= 0x02,
    ROM_ERASE_4K			= 0x20,
    ROM_ERASE_32K			= 0x52,
    ROM_ERASE_64K			= 0xd8,
    ROM_ERASE_ALL			= 0xc7,
    ROM_READ_SR1			= 0x05,
    ROM_WRITE_SR1			= 0x01,
    ROM_READ_SR2			= 0x35,
    ROM_WRITE_SR2			= 0x31,
    ROM_READ_SR3			= 0x15,
    ROM_WRITE_SR3			= 0x11,
    ROM_READ_SFPD			= 0x5a,
    ROM_ERASE_SECURITY			= 0x44,
    ROM_PROGRAM_SECURITY		= 0x42,
    ROM_READ_SECURITY			= 0x48,
    ROM_GLOBAL_BLOCK_LOCK		= 0x7e,
    ROM_GLOBAL_BLOCK_UNLOCK		= 0x98,
    ROM_READ_BLOCK_LOCK			= 0x3d,
    ROM_ONE_BLOCK_LOCK			= 0x36,
    ROM_ONE_BLOCK_UNLOCK		= 0x39,
    ROM_ERASE_PROGRAM_SUSPEND		= 0x75,
    ROM_ERASE_PROGRAM_RESUME		= 0x7a,
    ROM_POWER_DOWN			= 0xb9,
    ROM_ENABLE_RESET			= 0x66,
    ROM_RESET				= 0x99,
    ROM_FAST_READ_DUAL			= 0x3b
};

void romcopy_download(void *dst, size_t offset, size_t len)
{
    ROMCOPY_RAMADDR = (size_t)dst;
    ROMCOPY_ROMCMD  = __rom_offset + offset + (ROM_FAST_READ_DUAL << 24);
    ROMCOPY_DATALEN = len | ROMCOPY_SPI_CMDLEN(5) | ROMCOPY_SPI_DUAL;
}

void romcopy_bzero(void *dst, size_t len)
{
    ROMCOPY_RAMADDR = (size_t)dst;
    ROMCOPY_DATALEN = len | ROMCOPY_ZERO_BUFFER;
}

uint32_t romcopy_time[2];
IRQHANDLER(romcopy)
{
    static unsigned int romcopy_state;

    switch (romcopy_state++) {
    case 0:
	/* Copy DRAM data */
	romcopy_download(__dram_init_start, 0, _dram_init_len());
	break;
    case 1:
	/* Zero .dram.bss */
	romcopy_time[0] = rdtime() - time_zero;
	romcopy_bzero(__dram_bss_start, _dram_bss_len());
	break;
    default:
	romcopy_time[1] = rdtime() - romcopy_time[0];
	mask_irq(ROMCOPY_IRQ);
	return;
    }
}

/* Base32 without I O Z S */
static const char base32[33] = "ABCDEFGHJKLMNPQRTUVWXY0123456789";
char serial_str[14];

/* Read unique serial number programmed into ROM */
uint64_t rom_get_serial(void)
{
    union {
	uint32_t l[2];
	uint64_t q;
    } o;

    waitfor(ROMCOPY_IRQ);

    ROMCOPY_ROMCMD  = ROM_READ_UNIQUE_ID << 24;
    ROMCOPY_DATALEN = ROMCOPY_SPI_CMDLEN(5) | ROMCOPY_SPI_MORE;
    waitfor(ROMCOPY_IRQ);

    ROMCOPY_DATALEN = ROMCOPY_SPI_CMDLEN(4) | ROMCOPY_SPI_MORE;
    waitfor(ROMCOPY_IRQ);

    o.l[1] = ROMCOPY_INPUT;
    ROMCOPY_DATALEN = ROMCOPY_SPI_CMDLEN(4);
    waitfor(ROMCOPY_IRQ);

    o.l[0] = ROMCOPY_INPUT;

    uint64_t v = o.q;
    char *p = serial_str;
    for (int i = 0; i < 14; i++) {
	*p++ = base32[v & 31];
	v >>= 5;
    }
    *p = '\0';

    if ( 1 )
	con_printf("ROM serial: %08x-%08x (%s)\n", o.l[1], o.l[0], serial_str);

    return o.q;
}
